# pylint: skip-file
import pytest
from hypothesis import given, assume
from hypothesis.strategies import integers
from suspect.fbbt.propagation import BoundsPropagationVisitor
import suspect.dag.expressions as dex
from suspect.interval import Interval
from tests.conftest import (
    PlaceholderExpression,
    bound_description_to_bound,
    mono_description_to_mono,
    coefficients,
    reals,
)


class PropagationContext:
    def __init__(self, bounds=None):
        if bounds is None:
            bounds = {}
        self._bounds = bounds

    def bounds(self, expr):
        return self._bounds[expr]

    def get_bounds(self, expr):
        return self._bounds.get(expr)

    def set_bounds(self, expr, value):
        self._bounds[expr] = value


@pytest.fixture
def visitor():
    return BoundsPropagationVisitor()


@given(reals(), reals())
def test_variable_bound(visitor, a, b):
    lb = min(a, b)
    ub = max(a, b)
    var = dex.Variable('x0', lower_bound=lb, upper_bound=ub)
    ctx = PropagationContext()
    visitor.visit(var, ctx)
    assert ctx.bounds(var) == Interval(lb, ub)


@given(reals())
def test_constant_bound(visitor, c):
    const = dex.Constant(c)
    ctx = PropagationContext()
    visitor.visit(const, ctx)
    assert ctx.bounds(const) == Interval(c, c)


@given(reals(), reals(), reals(), reals())
def test_constraint_bound(visitor, a, b, c, d):
    e_lb, e_ub = min(a, b), max(a, b)
    c_lb, c_ub = min(c, d), max(c, d)
    assume(max(e_lb, c_lb) <= min(e_ub, c_ub))
    p = PlaceholderExpression()
    cons = dex.Constraint('c0', lower_bound=c_lb, upper_bound=c_ub, children=[p])
    ctx = PropagationContext({p: Interval(e_lb, e_ub)})
    visitor.visit(cons, ctx)
    expected = Interval(max(e_lb, c_lb), min(c_ub, e_ub))
    assert ctx.bounds(cons) == expected


@given(reals(max_value=100.0), reals(max_value=100.0))
def test_objective_bound(visitor, a, b):
    lb, ub = min(a, b), max(a, b)
    assume(lb < ub)
    p = PlaceholderExpression()
    o = dex.Objective('obj', children=[p])
    ctx = PropagationContext({p: Interval(lb, ub)})
    visitor.visit(o, ctx)
    assert ctx.bounds(o) == Interval(lb, ub)


@given(integers(min_value=1))
def test_even_pow_bound(visitor, ctx, n):
    p = PlaceholderExpression()
    c = dex.Constant(2*n)
    pow_ = dex.PowExpression(children=[p, c])
    ctx = PropagationContext()
    visitor.visit(pow_, ctx)
    assert ctx.bounds(pow_).is_nonnegative()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from numbers import Number
from pyomo.core.base import Var
from pyomo.core.base.var import (_GeneralVarData, SimpleVar)
import pyomo.core.base.expr_pyomo4 as omo


Variable = Var
NumericConstant = omo.NumericConstant
EqualityExpression = omo._EqualityExpression
InequalityExpression = omo._InequalityExpression
ProductExpression = omo._ProductExpression
DivisionExpression = omo._DivisionExpression
SumExpression = omo._SumExpression
LinearExpression = omo._LinearExpression
NegationExpression = omo._NegationExpression
UnaryFunctionExpression = omo._UnaryFunctionExpression
AbsExpression = omo._AbsExpression
PowExpression = omo._PowExpression


CALLBACKS_NAMES = {
    _GeneralVarData: 'visit_variable',
    SimpleVar: 'visit_variable',
    Variable: 'visit_variable',
    int: 'visit_number',
    float: 'visit_number',
    Number: 'visit_number',
    NumericConstant: 'visit_numeric_constant',
    EqualityExpression: 'visit_equality',
    InequalityExpression: 'visit_inequality',
    ProductExpression: 'visit_product',
    DivisionExpression: 'visit_division',
    SumExpression: 'visit_sum',
    LinearExpression: 'visit_linear',
    NegationExpression: 'visit_negation',
    UnaryFunctionExpression: 'visit_unary_function',
    AbsExpression: 'visit_abs',
    PowExpression: 'visit_pow',

}


def try_callback(handler, expr):
    """Try calling one of the registered callbacks, raising an exception
    if no callback matches.
    """
    callback_name = CALLBACKS_NAMES.get(type(expr))
    if callback_name is None:
        msg = 'No callback found for {} {}'.format(
            expr, type(expr)
        )
        raise RuntimeError(msg)
    else:
        callback = getattr(handler, callback_name)
        if callback is None:
            raise RuntimeError('Missing callback {}'.format(callback_name))
        callback(expr)


class ExpressionHandler(ABC):
    @abstractmethod
    def visit_number(self, n):
        pass

    @abstractmethod
    def visit_numeric_constant(self, expr):
        pass

    @abstractmethod
    def visit_variable(self, expr):
        pass

    @abstractmethod
    def visit_equality(self, expr):
        pass

    @abstractmethod
    def visit_inequality(self, expr):
        pass

    @abstractmethod
    def visit_product(self, expr):
        pass

    @abstractmethod
    def visit_division(self, expr):
        pass

    @abstractmethod
    def visit_sum(self, expr):
        pass

    @abstractmethod
    def visit_linear(self, expr):
        pass

    @abstractmethod
    def visit_negation(self, expr):
        pass

    @abstractmethod
    def visit_unary_function(self, expr):
        pass

    @abstractmethod
    def visit_abs(self, expr):
        pass

    @abstractmethod
    def visit_pow(self, expr):
        pass


def bottom_up_visit(handler, expr):
    """Visit the expression from leaf nodes to root.

    Parameters
    ----------
    handler: ExpressionHandler
        handler that will handle each sub-expression callback
    expr:
        the expression to visit
    """
    expr_level = {}
    expr_by_id = {}
    with omo.bypass_clone_check():
        stack = [(0, expr)]
        while len(stack) > 0:
            (lvl, expr) = stack.pop()

            old_lvl = expr_level.get(id(expr), -1)
            expr_level[id(expr)] = max(old_lvl, lvl)
            expr_by_id[id(expr)] = expr

            if isinstance(expr, omo._ExpressionBase):
                for arg in expr._args:
                    stack.append((lvl+1, arg))

        expr_level = sorted(
            [(lvl, ex) for ex, lvl in expr_level.items()],
            reverse=True,
        )

        for _, expr_id in expr_level:
            expr = expr_by_id[expr_id]
            try_callback(handler, expr)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import suspect.dag.expressions as dex
from io import StringIO


_NODE_SYMBOLS = {
    dex.ProductExpression: '*',
    dex.DivisionExpression: '/',
    dex.SumExpression: '+',
    dex.PowExpression: '^',
    dex.LinearExpression: 'Σ',
    dex.NegationExpression: '-',
    dex.AbsExpression: '|.|',
    dex.SqrtExpression: '√',
    dex.ExpExpression: 'exp',
    dex.LogExpression: 'log',
    dex.SinExpression: 'sin',
    dex.CosExpression: 'cos',
    dex.TanExpression: 'tan',
    dex.AsinExpression: 'asin',
    dex.AcosExpression: 'acos',
    dex.AtanExpression: 'atan',
}


def _node_name(v):
    return 'n{}'.format(id(v))


def _node_label(expr):
    if isinstance(expr, (dex.Variable, dex.Constraint, dex.Objective)):
        return expr.name
    elif isinstance(expr, dex.Constant):
        return str(expr.lower_bound)
    else:
        return _NODE_SYMBOLS.get(type(expr))


def dump(dag, f, metadata=None):
    f.write('digraph G {\n')
    f.write('  rankdir = TB;\n')
    f.write('  rank = same;\n')

    f.write('  { rank=same;\n')
    for cons in dag.constraints.values():
        f.write('    {} [label="{}", shape=box];\n'.format(
            _node_name(cons),
            _node_label(cons),
        ))

    for obj in dag.objectives.values():
        f.write('    {} [label="{}", shape=box, color=red];\n'.format(
            _node_name(obj),
            _node_label(obj),
        ))
    f.write('  }\n')

    cur_depth = 0
    f.write('  { rank=same;\n')
    for vertex in dag.vertices:
        if isinstance(vertex, (dex.Constraint, dex.Objective)):
            continue
        if vertex.depth != cur_depth:
            cur_depth += 1
            f.write('  }\n  { rankdir=same;\n')

        if metadata is None or metadata.get(vertex) is None:
            label = _node_label(vertex)
        elif metadata is not None and metadata.get(vertex) is not None:
            label = _node_label(vertex) + '\\n' + str(metadata[vertex])
        f.write('    {} [label="{}"];\n'.format(
            _node_name(vertex),
            label,
        ))
    f.write('  }\n')

    for vertex in dag.vertices:
        for i, child in enumerate(vertex.children):
            f.write('  {} -> {} [taillabel="{}"];\n'.format(
                _node_name(vertex),
                _node_name(child),
                i,
            ))

    f.write('}\n')


def dumps(dag, metadata=None):
    s = StringIO()
    dump(dag, s, metadata)
    return s.getvalue()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visitor implementation for ProblemDag."""
import abc
from suspect.util import deprecated
from suspect.interfaces import Visitor as IVisitor
from suspect.dag.expressions import Expression


class Visitor(IVisitor):
    """Visitor over vertices of ProblemDag."""
    def __init__(self, *_args):
        try:
            self._registered_callbacks = self.register_callbacks()
        except NotImplementedError:
            self._registered_callbacks = self.register_handlers()

    # @abc.abstractmethod
    def register_callbacks(self):
        """Register callbacks for each expression type."""        
        raise NotImplementedError()

    @deprecated('register_callbacks')
    def register_handlers(self):
        """Register callbacks for each expression type.

        DEPRECATED: use register_callbacks instead.
        """
        pass

    @abc.abstractmethod
    def handle_result(self, expr, result, ctx):
        """Handle visit result."""
        pass

    def _handle_result(self, expr, result, ctx):
        return self.handle_result(expr, result, ctx)

    def _visit_expr(self, expr, ctx, cb):
        result = cb(expr, ctx)
        return self._handle_result(expr, result, ctx)

    def visit(self, vertex, ctx): # pylint: disable=missing-docstring
        expr = vertex
        type_ = type(expr)
        cb = self._registered_callbacks.get(type_)
        if cb is not None:
            return self._visit_expr(expr, ctx, cb)

        # try superclasses, for most cases this will work fine
        # but since dicts are not ordered it could cause
        # unexpected behaviour
        for target_type, cb in self._registered_callbacks.items():
            if isinstance(expr, target_type):
                return self._visit_expr(expr, ctx, cb)
        return True


class ForwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting ProblemDag forward."""
    pass


class BackwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting ProblemDag backward."""
    def _handle_result(self, _expr, result, ctx):
        if result is None:
            return False
        any_change = False
        for child, value in result.items():
            any_change |= self.handle_result(child, value, ctx)
        return any_change



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to run bound tightener or a problem."""
from suspect.fbbt.initialization import BoundsInitializationVisitor
from suspect.fbbt.propagation import BoundsPropagationVisitor
from suspect.fbbt.tightening import BoundsTighteningVisitor


class FBBTStopCriterion(object):
    """Stop criterion for FBBT."""
    def __init__(self, max_iter=10):
        self._max_iter = max_iter
        self._iter = 0

    def should_stop(self):
        """Predicate to check if FBBT should stop."""
        return self._iter >= self._max_iter

    def iteration_end(self):
        """Callback called at the end of each FBBT iteration."""
        self._iter += 1

    def intercept_changes(self, visitor):
        if not hasattr(visitor, 'handle_result'):
            raise RuntimeError('expected decorated visitor to have handle_result method')
        original_visitor_handle = visitor.handle_result
        def _wrapper(expr, new_bound, ctx):
            return original_visitor_handle(expr, new_bound, ctx)
        visitor.handle_result = _wrapper
        return visitor


class BoundsTightener(object):
    """Configure and run FBBT on a problem.

    Parameters
    ----------
    forward_iterator:
       forward iterator over vertices of the problem
    backward_iterator:
       backward iterator over vertices of the problem
    stop_criterion:
       criterion used to stop iteration
    """
    def __init__(self, forward_iterator, backward_iterator, stop_criterion):
        self._forward_iterator = forward_iterator
        self._backward_iterator = backward_iterator
        self._stop_criterion = stop_criterion

    def tighten(self, problem, ctx):
        """Tighten bounds of ``problem`` storing them in ``ctx``."""
        self._forward_iterator.iterate(problem, BoundsInitializationVisitor(), ctx)
        prop_visitor = self._stop_criterion.intercept_changes(BoundsPropagationVisitor())
        tigh_visitor = self._stop_criterion.intercept_changes(BoundsTighteningVisitor())
        changes_tigh = None
        changes_prop = None
        while not self._stop_criterion.should_stop():
            changes_prop = self._forward_iterator.iterate(
                problem, prop_visitor, ctx, starting_vertices=changes_tigh
            )
            changes_tigh = self._backward_iterator.iterate(
                problem, tigh_visitor, ctx, starting_vertices=changes_prop
            )
            self._stop_criterion.iteration_end()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds initialization visitor."""
from suspect.interval import Interval
from suspect.interfaces import CombineUnaryFunctionRules
from suspect.visitor import BackwardVisitor
from suspect.fbbt.initialization.rules import (
    SqrtRule,
    LogRule,
    AsinRule,
    AcosRule,
)


class BoundsInitializationVisitor(BackwardVisitor):
    """Initialize problem bounds using function domains as bound."""
    needs_matching_rules = False

    def register_rules(self):
        return [
            CombineUnaryFunctionRules(
                SqrtRule(),
                LogRule(),
                AsinRule(),
                AcosRule(),
            )
        ]

    def handle_result(self, expr, value, ctx):
        if value is None:
            new_bounds = Interval(None, None)
        else:
            new_bounds = value
        ctx.set_bounds(expr, new_bounds)
        return True



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for product expressions."""
from suspect.convexity.convexity import Convexity
from suspect.expression import ExpressionType
from suspect.interfaces import Rule


class ProductRule(Rule):
    """Return convexity of product."""
    root_expr = ExpressionType.Product

    def apply(self, expr, ctx):
        f, g = expr.children
        if f is g:
            return _product_as_square_convexity(f, ctx)

        # -c*x*x is encoded as Linear([-c], [x])*Variable(x)
        if f.expression_type == ExpressionType.Linear and g.expression_type == ExpressionType.Variable:
            cvx = _product_linear_by_variable_convexity(f, g)
            if cvx is not None:
                return cvx

        if g.expression_type == ExpressionType.Linear and f.expression_type == ExpressionType.Variable:
            cvx = _product_linear_by_variable_convexity(g, f)
            if cvx is not None:
                return cvx

        mono_g = ctx.monotonicity(g)
        mono_f = ctx.monotonicity(f)
        if mono_g.is_constant():
            return _product_convexity(f, g, ctx)
        elif mono_f.is_constant():
            return _product_convexity(g, f, ctx)
        return Convexity.Unknown


def _product_as_square_convexity(f, ctx):
    # same as f**2
    cvx_f = ctx.convexity(f)
    bounds_f = ctx.bounds(f)
    if cvx_f.is_linear():
        return Convexity.Convex
    elif cvx_f.is_convex() and bounds_f.is_nonnegative():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_f.is_nonpositive():
        return Convexity.Convex
    return Convexity.Unknown


def _product_linear_by_variable_convexity(f, g):
    if len(f.children) == 1 and f.children[0] is g:
        if f.coefficients[0] > 0:
            return Convexity.Convex
        return Convexity.Concave
    return None


def _product_convexity(f, g, ctx):
    cvx_f = ctx.convexity(f)
    bounds_g = ctx.bounds(g)

    if cvx_f.is_linear():
        return cvx_f
    elif cvx_f.is_convex() and bounds_g.is_nonnegative():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_nonpositive():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_nonnegative():
        return Convexity.Concave
    elif cvx_f.is_convex() and bounds_g.is_nonpositive():
        return Convexity.Concave
    return Convexity.Unknown



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visitor applying rules for polynomial degree computation."""
from suspect.visitor import Visitor
from suspect.polynomial.rules import * # pylint: disable=wildcard-import


class PolynomialDegreeVisitor(Visitor):
    """Visitor applying polynomiality rules."""
    def handle_result(self, expr, result, ctx):
        ctx[expr] = result
        return True

    def register_rules(self):
        return [
            VariableRule(),
            ConstantRule(),
            ConstraintRule(),
            ObjectiveRule(),
            DivisionRule(),
            ProductRule(),
            LinearRule(),
            SumRule(),
            NegationRule(),
            PowerRule(),
            UnaryFunctionRule(),
        ]



# pylint: skip-file
import pytest
from tests.conftest import PlaceholderExpression
from suspect.dag.dag import VerticesList
from hypothesis import given, assume
import hypothesis.strategies as st


class TestVerticesList(object):
    @pytest.mark.parametrize('reverse', [True, False])
    @given(depths=st.lists(st.integers(min_value=0, max_value=100), min_size=1))
    def test_with_starting_vertices(self, depths, reverse):
        vertices = [PlaceholderExpression() for _ in depths]
        for i, v in enumerate(vertices):
            v.depth = depths[i]
        vl = VerticesList(vertices, reverse=reverse)
        assert [v.depth for v in vl] == sorted(depths, reverse=reverse)

    @pytest.mark.parametrize('reverse', [True, False])
    @given(depths=st.lists(st.integers(min_value=0, max_value=100), min_size=1))
    def test_append(self, depths, reverse):
        vl = VerticesList(reverse=reverse)
        for d in depths:
            p = PlaceholderExpression()
            p.depth = d
            vl.append(p)
        assert [v.depth for v in vl] == sorted(depths, reverse=reverse)

    @pytest.mark.parametrize('reverse', [True, False])
    @given(depths=st.lists(st.integers(min_value=0, max_value=100), min_size=1))
    def test_pop(self, depths, reverse):
        vertices = [PlaceholderExpression() for _ in depths]
        for i, v in enumerate(vertices):
            v.depth = depths[i]
        vl = VerticesList(vertices, reverse=reverse)
        if reverse:
            assert vl.pop().depth == max(depths)
        else:
            assert vl.pop().depth == min(depths)

    def test_bool(self):
        vl = VerticesList()
        assert not vl
        vl.append(PlaceholderExpression())
        assert vl
        vl.pop()
        assert not vl



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for nondecreasing functions."""
from suspect.expression import UnaryFunctionType
from suspect.interfaces import UnaryFunctionRule


class NondecreasingFunctionRule(UnaryFunctionRule):
    """Return monotonicity of nondecreasing function."""
    def apply(self, expr, ctx):
        child = expr.children[0]
        mono = ctx.monotonicity(child)
        return mono


class SqrtRule(NondecreasingFunctionRule):
    """Return monotonicity of sqrt function."""
    func_type = UnaryFunctionType.Sqrt


class ExpRule(NondecreasingFunctionRule):
    """Return monotonicity of exp function."""
    func_type = UnaryFunctionType.Exp


class LogRule(NondecreasingFunctionRule):
    """Return monotonicity of log function."""
    func_type = UnaryFunctionType.Log


class TanRule(NondecreasingFunctionRule):
    """Return monotonicity of tan function."""
    func_type = UnaryFunctionType.Tan


class AsinRule(NondecreasingFunctionRule):
    """Return monotonicity of asin function."""
    func_type = UnaryFunctionType.Asin


class AtanRule(NondecreasingFunctionRule):
    """Return monotonicity of atan function."""
    func_type = UnaryFunctionType.Atan



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import abc
from suspect.expression import ExpressionType, UnaryFunctionType
from suspect.math import inf # pylint: disable=no-name-in-module


class Domain(Enum):
    """The variable domain"""
    REALS = 0
    INTEGERS = 1
    BINARY = 2


class Sense(Enum):
    """The objective function sense"""
    MINIMIZE = 0
    MAXIMIZE = 1


class Expression(metaclass=abc.ABCMeta):
    """The base class for all expressions objects in the DAG"""
    is_source = False
    is_sink = False

    def __init__(self, children=None):
        if children is None:
            children = []

        self._children = children
        self._parents = []

        self._depth = 0
        self._update_depth()

    @property
    def depth(self):
        """The depth of the expression.

        The depth of the expression is defined as `0` if the
        expression is a source (Variables and Constants), otherwise
        it is the maximum depth of its children plus `1`.
        """
        return self._depth

    @property
    def children(self):
        return self._children

    @property
    def parents(self):
        return self._parents

    def add_parent(self, parent):
        self._parents.append(parent)

    def _update_depth(self):
        max_depth = self.depth
        for child in self.children:
            if child.depth >= max_depth:
                max_depth = child.depth + 1
        self._depth = max_depth

    def is_constant(self):
        return False

    def __hash__(self):
        return hash(id(self))

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not (self == other)


class ProductExpression(Expression):
    expression_type = ExpressionType.Product


class DivisionExpression(Expression):
    expression_type = ExpressionType.Division


class SumExpression(Expression):
    expression_type = ExpressionType.Sum


class PowExpression(Expression):
    expression_type = ExpressionType.Power


class LinearExpression(Expression):
    expression_type = ExpressionType.Linear

    def __init__(self, coefficients=None, children=None, constant_term=None):
        super().__init__(children)
        if coefficients is None:
            coefficients = []
        if constant_term is None:
            constant_term = 0.0
        self.constant_term = constant_term
        self.coefficients = coefficients
        self._check_coefficients()

    def _check_coefficients(self):
        assert len(self.coefficients) == len(self.children)


class UnaryFunctionExpression(Expression):
    expression_type = ExpressionType.UnaryFunction
    def __init__(self, children=None):
        super().__init__(children)
        assert len(self.children) == 1


class NegationExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.Negation
    func_name = 'negation'


class AbsExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Abs
    func_name = 'abs'


class SqrtExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Sqrt
    func_name = 'sqrt'


class ExpExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Exp
    func_name = 'exp'


class LogExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Log
    func_name = 'log'


class SinExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Sin
    func_name = 'sin'


class CosExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Cos
    func_name = 'cos'


class TanExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Tan
    func_name = 'tan'


class AsinExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Asin
    func_name = 'asin'


class AcosExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Acos
    func_name = 'acos'


class AtanExpression(UnaryFunctionExpression):
    expression_type = ExpressionType.UnaryFunction
    func_type = UnaryFunctionType.Atan
    func_name = 'atan'


class Objective(Expression):
    expression_type = ExpressionType.Objective
    is_sink = True

    def __init__(self, name, sense=None, children=None):
        super().__init__(children)
        if sense is None:
            sense = Sense.MINIMIZE
        self.sense = sense
        self.name = name

    def is_minimizing(self):
        return self.sense == Sense.MINIMIZE

    def is_maximizing(self):
        return self.sense == Sense.MAXIMIZE


class BoundedExpression(Expression):
    def __init__(self, lower_bound, upper_bound, children=None):
        super().__init__(children)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def bounded_below(self):
        return self.lower_bound is not None and self.lower_bound != -inf

    def bounded_above(self):
        return self.upper_bound is not None and self.upper_bound != inf


class Constraint(BoundedExpression):
    expression_type = ExpressionType.Constraint
    is_sink = True

    def __init__(self, name, lower_bound, upper_bound, children=None):
        super().__init__(lower_bound, upper_bound, children)
        self.name = name

    def linear_component(self):
        linear, _ = self._split_linear_nonlinear()
        return linear

    def nonlinear_component(self):
        _, nonlinear = self._split_linear_nonlinear()
        return nonlinear

    def _split_linear_nonlinear(self):
        child = self.children[0]

        if isinstance(child, LinearExpression):
            return child, None

        if not isinstance(child, SumExpression):
            return None, [child]

        linear = None
        nonlinear = []
        for arg in child.children:
            if isinstance(arg, LinearExpression):
                if linear is not None:
                    raise AssertionError(
                        'Constraint root should have only one LinearExpression child'
                    )
                linear = arg
            else:
                nonlinear.append(arg)
        return linear, nonlinear

    def is_equality(self):
        return self.lower_bound == self.upper_bound

    def __str__(self):
        return 'Constraint(name={}, lower_bound={}, upper_bound={}, children={})'.format(
            self.name, self.lower_bound, self.upper_bound, self.children
        )


class Variable(BoundedExpression):
    expression_type = ExpressionType.Variable
    is_source = True

    def __init__(self, name, lower_bound, upper_bound, domain=None):
        super().__init__(lower_bound, upper_bound, None)
        self.domain = domain
        self.name = name

    def is_binary(self):
        return self.domain == Domain.BINARY

    def is_integer(self):
        return self.domain == Domain.INTEGERS

    def is_real(self):
        return self.domain == Domain.REALS

    def is_constant(self):
        # TODO(fracek): if we ever support fixing variables, change this
        return False

    def __str__(self):
        return 'Variable(name={}, lower_bound={}, upper_bound={}, domain={})'.format(
            self.name, self.lower_bound, self.upper_bound, self.domain
        )


class Constant(BoundedExpression):
    expression_type = ExpressionType.Constant
    is_source = True

    def __init__(self, value):
        super().__init__(value, value, None)

    @property
    def value(self):
        assert self.lower_bound == self.upper_bound
        return self.lower_bound

    def is_constant(self):
        return True



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyomo.core.base.expr import set_expression_tree_format
import pyomo.core.base.expr_common as common


def set_pyomo4_expression_tree():
    """Set Pyomo expression tree format to ``Mode.pyomo4_trees``.

    SUSPECT does not work with Pyomo default tree format, so this
    function should be called at the beginning of every program.
    """
    set_expression_tree_format(common.Mode.pyomo4_trees)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from suspect.dag.iterator import DagForwardIterator
from suspect.polynomial import PolynomialDegreeVisitor


def polynomial_degree(dag, ctx):
    """Compute polynomial degree of expressions in dag."""
    iterator = DagForwardIterator()
    iterator.iterate(dag, PolynomialDegreeVisitor(), ctx)
    return ctx



# pylint: skip-file
import pytest
from tests.conftest import PlaceholderExpression
from suspect.dag.dag import ProblemDag
from suspect.dag.iterator import DagForwardIterator, DagBackwardIterator
import suspect.dag.expressions as dex
from hypothesis import given, assume
import hypothesis.strategies as st


@pytest.fixture
def problem_dag():
    dag = ProblemDag()

    # build a dag that looks like this:
    #
    #            c0    c1
    #            |     |
    #            |     |
    #            +    |.|
    #            | \   |
    #            |   \ |
    #            S     ^
    #          /  \   / \
    #        x0     x1   c

    x0 = dex.Variable('x0', None, None)
    dag.add_variable(x0)
    x1 = dex.Variable('x1', None, None)
    dag.add_variable(x1)
    c = dex.Constant(1.0)
    dag.add_vertex(c)

    n0 = dex.LinearExpression([1.0, -1.0], [x0, x1])
    x0.add_parent(n0)
    x1.add_parent(n0)
    dag.add_vertex(n0)

    n1 = dex.PowExpression([x1, c])
    x1.add_parent(n1)
    c.add_parent(n1)
    dag.add_vertex(n1)

    n2 = dex.SumExpression([n0, n1])
    n0.add_parent(n2)
    n1.add_parent(n2)
    dag.add_vertex(n2)

    n3 = dex.AbsExpression([n1])
    n1.add_parent(n3)
    dag.add_vertex(n3)

    c0 = dex.Constraint('c0', None, None, [n2])
    n2.add_parent(c0)
    dag.add_constraint(c0)

    c1 = dex.Constraint('c1', None, None, [n3])
    n3.add_parent(c1)
    dag.add_constraint(c1)

    return dag


class FakeForwardVisitor(object):
    def __init__(self, visit_returns):
        self.seen = set()
        self._visit_returns = visit_returns

    def visit(self, expr, ctx):
        self.seen.add(expr)
        return self._visit_returns(expr)


class TestDagForwardIterator(object):
    def test_no_starting_point(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(lambda v: True)
        changes = it.iterate(problem_dag, visitor, None)
        assert changes[0].is_source
        assert changes[-1].is_sink
        assert len(changes) == 9

    def test_early_stop(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(
            lambda v: False if isinstance(v, dex.PowExpression) else True,
        )
        changes = it.iterate(problem_dag, visitor, None)
        assert len(changes) == 9-3

    def test_starting_point(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(lambda v: False)
        changes = it.iterate(problem_dag, visitor, None, starting_vertices=[])
        assert len(visitor.seen) == 3
        assert len(changes) == 0


class FakeBackwardVisitor(object):
    def __init__(self, visit_returns):
        self.seen = set()
        self._visit_returns = visit_returns

    def visit(self, expr, ctx):
        self.seen.add(expr)
        return self._visit_returns(expr)


class TestDagBackwardIterator(object):
    def test_no_starting_point(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(lambda v: True)
        changes = it.iterate(problem_dag, visitor, None)
        assert changes[0].is_sink
        assert changes[-1].is_source
        assert len(changes) == 9

    def test_early_stop(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(
            lambda v: False if isinstance(v, dex.SumExpression) else True,
        )
        changes = it.iterate(problem_dag, visitor, None)
        assert len(changes) == 9-3

    def test_starting_point(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(lambda v: False)
        changes = it.iterate(problem_dag, visitor, None, starting_vertices=[])
        assert len(visitor.seen) == 2
        assert len(changes) == 0



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds tightening rules."""
from suspect.interfaces import Rule, UnaryFunctionRule
from suspect.expression import ExpressionType, UnaryFunctionType
from suspect.interval import Interval
from suspect.math import almosteq, inf # pylint: disable=no-name-in-module


MAX_EXPR_CHILDREN = 1000


class ConstraintRule(Rule):
    """Return new bounds for constraint."""
    root_expr = ExpressionType.Constraint

    def apply(self, expr, ctx):
        child = expr.children[0]
        bounds = Interval(expr.lower_bound, expr.upper_bound)
        return {
            child: bounds
        }


class SumRule(Rule):
    """Return new bounds for sum."""
    root_expr = ExpressionType.Sum

    def apply(self, expr, ctx):
        expr_bound = ctx.bounds(expr)
        if expr_bound.size() == inf:
            return None
        if len(expr.children) > MAX_EXPR_CHILDREN: # pragma: no cover
            return None
        child_bounds = {}
        for child, siblings in _sum_child_and_siblings(expr.children):
            siblings_bound = sum(ctx.bounds(s) for s in siblings)
            child_bounds[child] = expr_bound - siblings_bound
        return child_bounds


class LinearRule(Rule):
    """Return new bounds for linear expressions."""
    root_expr = ExpressionType.Linear

    def apply(self, expr, ctx):
        expr_bound = ctx.bounds(expr)
        if expr_bound.size() == inf:
            return None
        if len(expr.children) > MAX_EXPR_CHILDREN: # pragma: no cover
            return None
        child_bounds = {}
        const = expr.constant_term
        for (child_c, child), siblings in _linear_child_and_siblings(expr.coefficients, expr.children):
            siblings_bound = sum(ctx.bounds(s) * c for c, s in siblings) + const
            child_bounds[child] = (expr_bound - siblings_bound) / child_c
        return child_bounds


class PowerRule(Rule):
    """Return new bounds for power expressions."""
    root_expr = ExpressionType.Power

    def apply(self, expr, ctx):
        base, expo = expr.children
        if not expo.is_constant():
            return None
        if not almosteq(expo.value, 2):
            return None

        bounds = ctx.bounds(expr)
        # the bound of a square number is never negative, but check anyway to
        # avoid unexpected crashes.
        if not bounds.is_nonnegative():
            return None

        sqrt_bound = bounds.sqrt()
        return {
            base: Interval(-sqrt_bound.upper_bound, sqrt_bound.upper_bound)
        }


class _UnaryFunctionRule(UnaryFunctionRule):
    def apply(self, expr, ctx):
        child = expr.children[0]
        bounds = ctx.bounds(expr)
        return {
            child: self._child_bounds(bounds)
        }

    def _child_bounds(self, bounds):
        pass


class _BoundedFunctionRule(_UnaryFunctionRule):
    func_name = None

    def _child_bounds(self, bounds):
        func = getattr(bounds, self.func_name)
        return func.inverse()


class AbsRule(_BoundedFunctionRule):
    """Return new bounds for abs."""
    func_name = 'abs'


class SqrtRule(_BoundedFunctionRule):
    """Return new bounds for sqrt."""
    func_name = 'sqrt'


class ExpRule(_BoundedFunctionRule):
    """Return new bounds for exp."""
    func_name = 'exp'


class LogRule(_BoundedFunctionRule):
    """Return new bounds for log."""
    func_name = 'log'


class _UnboundedFunctionRule(_UnaryFunctionRule):
    def _child_bounds(self, bounds):
        return Interval(None, None)


class SinRule(_UnboundedFunctionRule):
    """Return new bounds for sin."""
    func_type = UnaryFunctionType.Sin


class CosRule(_UnboundedFunctionRule):
    """Return new bounds for cos."""
    func_type = UnaryFunctionType.Cos


class TanRule(_UnboundedFunctionRule):
    """Return new bounds for tan."""
    func_type = UnaryFunctionType.Tan


class AsinRule(_UnboundedFunctionRule):
    """Return new bounds for asin."""
    func_type = UnaryFunctionType.Asin


class AcosRule(_UnboundedFunctionRule):
    """Return new bounds for acos."""
    func_type = UnaryFunctionType.Acos


class AtanRule(_UnboundedFunctionRule):
    """Return new bounds for atan."""
    func_type = UnaryFunctionType.Atan


def _sum_child_and_siblings(children):
    for i, _ in enumerate(children):
        yield children[i], children[:i] + children[i+1:]


def _linear_child_and_siblings(coefficients, children):
    for i, child in enumerate(children):
        child_c = coefficients[i]
        other_children = children[:i] + children[i+1:]
        other_coefficients = coefficients[:i] + coefficients[i+1:]
        yield (child_c, child), zip(other_coefficients, other_children)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for visitors applying the given rules."""
import abc


class Visitor(metaclass=abc.ABCMeta):
    """Base class for visitors applying rules to DAG vertices."""
    needs_matching_rules = True

    def __init__(self):
        self._rules = self.register_rules()
        self._callbacks = self._init_callbacks()

    @abc.abstractmethod
    def register_rules(self):
        """Register rules applied by this visitor."""
        pass

    @abc.abstractmethod
    def handle_result(self, expr, result, ctx):
        """Handle ``result`` obtained applying rule to ``expr``."""
        pass

    def visit(self, expr, ctx):
        """Apply registered rule to ``expr``."""
        callback = self._callbacks.get(expr.expression_type)
        if callback is not None:
            return self._visit_expression(expr, ctx, callback)
        if self.needs_matching_rules:
            raise RuntimeError('visiting expression with no rule associated.')
        return False

    def _init_callbacks(self):
        callbacks = {}
        for rule in self._rules:
            callbacks[rule.root_expr] = rule.apply
        return callbacks

    def _handle_result(self, expr, result, ctx):
        return self.handle_result(expr, result, ctx)

    def _visit_expression(self, expr, ctx, callback):
        result = callback(expr, ctx)
        return self._handle_result(expr, result, ctx)


class ForwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting DAG forward."""
    pass


class BackwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting DAG backward."""
    def _handle_result(self, _expr, result, ctx):
        if result is None:
            # if we allow matching rules, then continue iteration
            return not self.needs_matching_rules
        any_change = False
        for child, value in result.items():
            any_change |= self.handle_result(child, value, ctx)
        return any_change



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity rules."""
from .base import VariableRule, ConstantRule, ConstraintRule, ObjectiveRule
from .abs import AbsRule
from .division import DivisionRule
from .linear import LinearRule, SumRule
from .concave import SqrtRule, LogRule
from .convex import ExpRule
from .negation import NegationRule
from .pow import PowerRule
from .product import ProductRule
from .trig import SinRule, CosRule, TanRule, AsinRule, AcosRule, AtanRule


__all__ = [
    'VariableRule', 'ConstantRule', 'ConstraintRule', 'ObjectiveRule',
    'ProductRule', 'DivisionRule', 'LinearRule', 'SumRule', 'AbsRule',
    'SqrtRule', 'ExpRule', 'LogRule', 'TanRule', 'AsinRule', 'AtanRule',
    'AcosRule', 'NegationRule', 'PowerRule', 'SinRule', 'CosRule',
]



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds propagation visitor."""

from suspect.visitor import ForwardVisitor
from suspect.fbbt.propagation.rules import (
    VariableRule,
    ConstantRule,
    ConstraintRule,
    ObjectiveRule,
    ProductRule,
    DivisionRule,
    LinearRule,
    SumRule,
    PowerRule,
    NegationRule,
    UnaryFunctionRule,
)


class BoundsPropagationVisitor(ForwardVisitor):
    """Tighten bounds from sources to sinks."""
    def register_rules(self):
        return [
            VariableRule(),
            ConstantRule(),
            ConstraintRule(),
            ObjectiveRule(),
            ProductRule(),
            DivisionRule(),
            LinearRule(),
            SumRule(),
            PowerRule(),
            NegationRule(),
            UnaryFunctionRule(),
        ]

    def handle_result(self, expr, new_bounds, ctx):
        old_bounds = ctx.get_bounds(expr)
        if old_bounds is not None:
            # bounds exists, tighten it
            new_bounds = old_bounds.intersect(new_bounds)
            has_changed = new_bounds != old_bounds
        else:
            has_changed = True

        ctx.set_bounds(expr, new_bounds)
        # return has_changed
        return True
